(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('d3-selection')) :
	typeof define === 'function' && define.amd ? define(['exports', 'd3-selection'], factory) :
	(factory((global.d3 = global.d3 || {}),global.d3));
}(this, (function (exports,d3Selection) { 'use strict';

var instanceLocal = d3Selection.local();
var noop = function (){}; // no operation

var component = function (tagName, className){
  var create = noop,
      render = noop,
      destroy = noop,
      key;

  function component(selection, data, context){
    var instances = (selection.nodeName ? d3Selection.select(selection) : selection)
      .selectAll(children)
      .filter(belongsToMe)
      .data(dataArray(data, context), key);
    instances
      .exit()
        .each(destroyInstance);
    return instances
      .enter().append(tagName)
        .attr("class", className)
        .each(createInstance)
      .merge(instances)
        .each(render);
  }

  function children(){
    return this.children;
  }

  function belongsToMe(){
    var instance = instanceLocal.get(this);
    return instance && instance.owner === component;
  }

  function dataArray(data, context){
    data = Array.isArray(data) ? data : [data];
    return context ? data.map(function (d){
      return Object.assign(Object.create(context), d);
    }) : data;
  }

  function createInstance(d, i, nodes){
    instanceLocal.set(this, {
      owner: component,
      destroy: function (){
        return destroy.call(this, d, i, nodes);
      }.bind(this)
    });
    create.call(this, d, i, nodes);
  }

  function destroyInstance(){
    d3Selection.select(this).selectAll("*").each(destroyDescendant);
    (instanceLocal.get(this).destroy() || d3Selection.select(this)).remove();
  }

  function destroyDescendant(){
    var instance = instanceLocal.get(this);
    instanceLocal.remove(this) && instance.destroy();
  }

  component.render = function(_) { return (render = _, component); };
  component.create = function(_) { return (create = _, component); };
  component.destroy = function(_) { return (destroy = _, component); };
  component.key = function(_) { return (key = _, component); };

  return component;
};

exports.component = component;

Object.defineProperty(exports, '__esModule', { value: true });

})));
